/*
	SuperCollider real time audio synthesis system
    Copyright (c) 2002 James McCartney. All rights reserved.
	http://www.audiosynth.com

    A count-leading-zeroes function, and those that can be derived from it.

    Adapted slightly by Andy Wingo.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __CLZ_H__
#define __CLZ_H__

#include <stdint.h>

#ifdef __MWERKS__
// powerpc native count leading zeroes instruction:
#define CLZ(x)  ((int)__cntlzw((unsigned int)x))
#elif defined (__ppc__)
static inline int CLZ(int arg) {
         __asm__ volatile("cntlzw %0, %1" : "=r" (arg) : "r" (arg));
         return arg;
}
#elif defined(__i386__)
static inline int CLZ(int arg) {
         __asm__ volatile("bsrl %0, %0\n\txorl $31,%0\n" : "=r" (arg) : "0" (arg));
         return arg;
}
#else
#error Implement me in C!
#endif

// count trailing zeroes
inline int32_t CTZ(int32_t x) 
{
	return 32 - CLZ(~x & (x-1));
}

// count leading ones
inline int32_t CLO(int32_t x) 
{
	return CLZ(~x);
}

// count trailing ones
inline int32_t CTO(int32_t x) 
{
	return 32 - CLZ(x & (~x-1));
}

// number of bits required to represent x. 
inline int32_t NUMBITS(int32_t x) 
{
	return 32 - CLZ(x);
}

// log2 of the next power of two greater than or equal to x. 
inline int32_t LOG2CEIL(int32_t x) 
{
	return 32 - CLZ(x - 1);
}

// next power of two greater than or equal to x
inline int32_t NEXTPOWEROFTWO(int32_t x) 
{
	return 1L << LOG2CEIL(x);
}

// is x a power of two
inline int ISPOWEROFTWO(int32_t x) 
{
	return (x & (x-1)) == 0;
}

// input a series of counting integers, outputs a series of gray codes .
inline int32_t GRAYCODE(int32_t x)
{
	return x & (x>>1);
}

// find least significant bit
inline int32_t LSBit(int32_t x)
{
	return x & -x;
}

// find least significant bit position
inline int32_t LSBitPos(int32_t x)
{
	return CTZ(x & -x);
}

// find most significant bit position
inline int32_t MSBitPos(int32_t x)
{
	return 31 - CLZ(x);
}

// find most significant bit
inline int32_t MSBit(int32_t x)
{
	return 1L << MSBitPos(x);
}

// count number of one bits
inline uint32_t ONES(uint32_t x) 
{
	uint32_t t; 
	x = x - ((x >> 1) & 0x55555555); 
	t = ((x >> 2) & 0x33333333); 
	x = (x & 0x33333333) + t; 
	x = (x + (x >> 4)) & 0x0F0F0F0F; 
	x = x + (x << 8); 
	x = x + (x << 16); 
	return x >> 24; 
}

// count number of zero bits
inline uint32_t ZEROES(uint32_t x)
{
	return ONES(~x);
}

#endif /* __CLZ_H__ */
